#ifdef OPENCV_C_HEADERS
#include <opencv2/core/types_c.h>
#include <opencv2/highgui/highgui_c.h>
#include <opencv2/imgproc/imgproc_c.h>
#include <opencv2/imgcodecs/legacy/constants_c.h>
#endif

#include <opencv2/core/core.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <opencv2/highgui/highgui.hpp>
#include "include/ASICamera2.h"
#include <sys/time.h>
#include <sys/stat.h>
#include <time.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <errno.h>
#include <string>
#include <iostream>
#include <cstdio>
#include <tr1/memory>
#include <ctime>
#include <stdlib.h>
#include <signal.h>
#include <fstream>

#define KNRM "\x1B[0m"
#define KRED "\x1B[31m"
#define KGRN "\x1B[32m"
#define KYEL "\x1B[33m"
#define KBLU "\x1B[34m"
#define KMAG "\x1B[35m"
#define KCYN "\x1B[36m"
#define KWHT "\x1B[37m"

#define USE_HISTOGRAM	// ac4lt
#define USE_NEW_EXPOSURE // ac4lt
#define US_IN_MS 1000                     // microseconds in a millisecond
#define MS_IN_SEC 1000                    // milliseconds in a second
#define US_IN_SEC (US_IN_MS * MS_IN_SEC)  // microseconds in a second

//-------------------------------------------------------------------------------------------------------
//-------------------------------------------------------------------------------------------------------

cv::Mat pRgb;
#define DEFAULT_FILENAME	"image.jpg"
char const *fileName = DEFAULT_FILENAME;
#define DEFAULT_TIMEFORMAT "%Y%m%d %H:%M:%S"
char timeFormat[50] = DEFAULT_TIMEFORMAT;    // ECC added
std::vector<int> compression_parameters;
bool bMain = true, bDisplay = false;
std::string dayOrNight;

bool bSaveRun = false, bSavingImg = false;
pthread_mutex_t mtx_SaveImg;
pthread_cond_t cond_SatrtSave;

int debugLevel = 0;
ASI_CONTROL_CAPS ControlCaps;   // ECC: made this global
int iNumOfCtrl = 0;             // ECC: made this global

// ECC added to ensure we don't try to update a non-updateable control, and to check for errors.
ASI_ERROR_CODE setControl(int CamNum, ASI_CONTROL_TYPE control, int value, ASI_BOOL makeAuto)
{
    ASI_ERROR_CODE ret = ASI_SUCCESS;
    int i;
    for (i = 0; i < iNumOfCtrl && i <= control; i++)    // controls are sorted 1 to n
    {
        ret = ASIGetControlCaps(CamNum, i, &ControlCaps);
        if (ControlCaps.ControlType == control)
        {
            if (ControlCaps.IsWritable)
            {
                if (value > ControlCaps.MaxValue)
                {
                    printf("WARNING: Value of %d greater than max value allowed (%ld) for control '%s'.\n", value, ControlCaps.MaxValue, ControlCaps.Name);
                    value = ControlCaps.MaxValue;
                } else if (value < ControlCaps.MinValue)
                {
                    printf("WARNING: Value of %d less than min value allowed (%ld) for control '%s'.\n", value, ControlCaps.MinValue, ControlCaps.Name);
                    value = ControlCaps.MinValue;
                }
                if (makeAuto == ASI_TRUE && ControlCaps.IsAutoSupported == ASI_FALSE)
                {
                    printf("WARNING: control '%s' doesn't support auto mode.\n", ControlCaps.Name);
                    makeAuto = ASI_FALSE;
                }
                ret = ASISetControlValue(CamNum, control, value, makeAuto);
            } else {
                printf("ERROR: ControlCap: '%s' not writable; not setting to %d.\n", ControlCaps.Name, value);
            }
            return ret;
        }
    }
    // printf("NOTICE: Camera does not support ControlCap # %d; not setting to %d.\n", control, value);
    return ASI_ERROR_INVALID_CONTROL_TYPE;
}

//-------------------------------------------------------------------------------------------------------
//-------------------------------------------------------------------------------------------------------

// Create Hex value from RGB
unsigned long createRGB(int r, int g, int b)
{
    return ((r & 0xff) << 16) + ((g & 0xff) << 8) + (b & 0xff);
}

void cvText(cv::Mat &img, const char *text, int x, int y, double fontsize, int linewidth, int linetype, int fontname,
            int fontcolor[], int imgtype, int outlinefont)
{
    if (imgtype == ASI_IMG_RAW16)
    {
	unsigned long fontcolor16 = createRGB(fontcolor[2], fontcolor[1], fontcolor[0]);
        if (outlinefont)
            cv::putText(img, text, cvPoint(x, y), fontname, fontsize, cvScalar(0,0,0), linewidth+4, linetype);
        cv::putText(img, text, cvPoint(x, y), fontname, fontsize, fontcolor16, linewidth, linetype);
    }
    else
    {
        if (outlinefont)
            cv::putText(img, text, cvPoint(x, y), fontname, fontsize, cvScalar(0,0,0, 255), linewidth+4, linetype);
        cv::putText(img, text, cvPoint(x, y), fontname, fontsize,
                    cvScalar(fontcolor[0], fontcolor[1], fontcolor[2], 255), linewidth, linetype);
    }
}

char *getTime()     // ECC: checking only for seconds is a bug, isn't it?  10:03:22 and 10:04:22 will match.
{                   // ECC: and why check at all?
    static int seconds_last = 99;
    static char TimeString[128];
    timeval curTime;
    gettimeofday(&curTime, NULL);
    if (seconds_last == curTime.tv_sec)
    {
        return 0;
    }

    seconds_last = curTime.tv_sec;
    strftime(TimeString, 80, timeFormat, localtime(&curTime.tv_sec));    // ECC: added "timeFormat"
    return TimeString;
}

char *getTime2()     // ECC added: this doesn't return 0 if seconds match
{
    //static int seconds_last = 99;
    static char TimeString[128];
    timeval curTime;
    gettimeofday(&curTime, NULL);
    //if (seconds_last == curTime.tv_sec)
    //{
    //    return 0;
    //}

    //seconds_last = curTime.tv_sec;
    strftime(TimeString, 80, timeFormat, localtime(&curTime.tv_sec));    // ECC: added "timeFormat"
    return TimeString;
}

std::string exec(const char *cmd)
{
    std::tr1::shared_ptr<FILE> pipe(popen(cmd, "r"), pclose);
    if (!pipe)
        return "ERROR";
    char buffer[128];
    std::string result = "";
    while (!feof(pipe.get()))
    {
        if (fgets(buffer, 128, pipe.get()) != NULL)
        {
            result += buffer;
        }
    }
    return result;
}

void *Display(void *params)     // ECC: xxxxxxxxxxxx this fails at cvShowImage().  When it fails, the thread appears to exit.
                                // Error message:   Unable to init server: Could not connect: Connection refused
                                // All calls to ASIGetVideoData() also fail with retcode=13 (buffer too small?).
{
    cv::Mat *pImg = (cv::Mat *)params;
    cvNamedWindow("video", 1);
    while (bDisplay)
    {
        cvShowImage("video", pImg);
        cvWaitKey(100);
    }
    cvDestroyWindow("video");
    printf("Display thread over\n");
    return (void *)0;
}

void displayDebugText(const char * text, int requiredLevel);    // ECC added to make compiler happy
void *SaveImgThd(void *para)
{
    while (bSaveRun)
    {
        char mybuf[256];  // ECC added
        pthread_mutex_lock(&mtx_SaveImg);
        pthread_cond_wait(&cond_SatrtSave, &mtx_SaveImg);
        bSavingImg = true;
        if (pRgb.data)
        {
            imwrite(fileName, pRgb, compression_parameters);
            if (dayOrNight == "NIGHT")
            {
                sprintf(mybuf, "scripts/saveImageNight.sh &");
            }
            else
            {
                sprintf(mybuf, "scripts/saveImageDay.sh &");
            }
            system(mybuf);  // ECC moved and changed
        } else {    // ECC added error message
            sprintf(mybuf, "----- SaveImgThd(): pRgb.data is null\n");
            displayDebugText(mybuf, 0);
        }
        bSavingImg = false;
        pthread_mutex_unlock(&mtx_SaveImg);
    }

    printf("save thread over\n");
    return (void *)0;
}

char retCodeBuffer[100];    // ECC added for debugging
char *getRetCode(ASI_ERROR_CODE code)
{
    std::string ret;
    if (code == ASI_SUCCESS) ret = "ASI_SUCCESS";
    else if (code == ASI_ERROR_INVALID_INDEX) ret = "ASI_ERROR_INVALID_INDEX";
	else if (code == ASI_ERROR_INVALID_ID) ret = "ASI_ERROR_INVALID_ID";
	else if (code == ASI_ERROR_INVALID_CONTROL_TYPE) ret = "ASI_ERROR_INVALID_CONTROL_TYPE";
	else if (code == ASI_ERROR_CAMERA_CLOSED) ret = "ASI_ERROR_CAMERA_CLOSED";
	else if (code == ASI_ERROR_CAMERA_REMOVED) ret = "ASI_ERROR_CAMERA_REMOVED";
	else if (code == ASI_ERROR_INVALID_PATH) ret = "ASI_ERROR_INVALID_PATH";
	else if (code == ASI_ERROR_INVALID_FILEFORMAT) ret = "ASI_ERROR_INVALID_FILEFORMAT";
	else if (code == ASI_ERROR_INVALID_SIZE) ret = "ASI_ERROR_INVALID_SIZE";
	else if (code == ASI_ERROR_INVALID_IMGTYPE) ret = "ASI_ERROR_INVALID_IMGTYPE";
	else if (code == ASI_ERROR_OUTOF_BOUNDARY) ret = "ASI_ERROR_OUTOF_BOUNDARY";
	else if (code == ASI_ERROR_TIMEOUT) ret = "ASI_ERROR_TIMEOUT";
	else if (code == ASI_ERROR_INVALID_SEQUENCE) ret = "ASI_ERROR_INVALID_SEQUENCE";
	else if (code == ASI_ERROR_BUFFER_TOO_SMALL) ret = "ASI_ERROR_BUFFER_TOO_SMALL";
	else if (code == ASI_ERROR_VIDEO_MODE_ACTIVE) ret = "ASI_ERROR_VIDEO_MODE_ACTIVE";
	else if (code == ASI_ERROR_EXPOSURE_IN_PROGRESS) ret = "ASI_ERROR_EXPOSURE_IN_PROGRESS";
	else if (code == ASI_ERROR_GENERAL_ERROR) ret = "ASI_ERROR_GENERAL_ERROR";
	else if (code == ASI_ERROR_END) ret = "ASI_ERROR_END";
	else if (code == -1) ret = "Non-ASI ERROR";
    else ret = "UNKNOWN ASI ERROR";

    sprintf(retCodeBuffer, "%d (%s)", (int) code, ret.c_str());
    return(retCodeBuffer);
}

int round10(int n)
{
    // Smaller multiple
    int a = (n / 10) * 10;

    // Larger multiple
    int b = a + 10;

    // Return of closest of two
    return (n - a > b - n)? b : a;
}

int bytesPerPixel(ASI_IMG_TYPE imageType) {
    int bytesPerPixel;
    switch (imageType) {
        case ASI_IMG_RGB24:
            bytesPerPixel = 3;
            break;
        case ASI_IMG_RAW16:
            bytesPerPixel = 2;
            break;
        case ASI_IMG_RAW8:
        case ASI_IMG_Y8:
        default:
            bytesPerPixel = 1;
    }
    return bytesPerPixel;
}

#ifdef USE_HISTOGRAM
// As a workaround to a bug in the ZWO SDK, where autoexposure daylight shots' exposures jump all over the place,
// this basically replaces ZWO's code with our auto-exposure code.
// We look at the brightness of an X by X square in image, and adjust exposure based on that.

int histogramBoxSize = 500;     // 500 px x 500 px box.  Must be a multiple of 2.
float histogramBoxPercentFromLeft = 0.5;   // % from left side that the center of the box is.  0.5 == the center of the image's X axis
float histogramBoxPercentFromTop = 0.5;    // % from top that the center of the box is.  0.5 == the center of the image's Y axis

void computeHistogram(unsigned char *imageBuffer, long width, long height, ASI_IMG_TYPE imageType, long *histogram) {
    long i ;
    unsigned char *b = imageBuffer;
    int h;

    for (h = 0; h < 256; h++) {
        histogram[h] = 0;
    }

    width *= bytesPerPixel(imageType);
    long roiX1 = (width * histogramBoxPercentFromLeft) - (histogramBoxSize / 2);
    long roiX2 = (width * histogramBoxPercentFromLeft) + (histogramBoxSize / 2);
    long roiY1 = (height * histogramBoxPercentFromTop) - (histogramBoxSize / 2);
    long roiY2 = (height * histogramBoxPercentFromTop) + (histogramBoxSize / 2);
    //printf("calculating histogram\n");
    for (long y = roiY1; y < roiY2; y++) {
        for (long x = roiX1; x < roiX2; x++) {
            i = (width * y) + x;
            //printf("x = %ld, y = %ld, i is %ld, b[i] is %i\n", x, y, i, b[i]);
            histogram[b[i]]++;
        }
    }
    //printf("calculation complete\n");
}

int calculateHistogramPeak(long *histogram) {
    int peakBin = 0;
    for (int h = 0; h < 256; h++) {
        //printf("%3d: %ld\n", h, histogram[h]);
        if (histogram[h] > histogram[peakBin])
            peakBin = h;
    }
    //printf("  > *** histogram peak is bin %d\n", peakBin);
    return peakBin;
}
#endif

long actualExposureMicroseconds = 0;	// actual exposure taken, per the camera
long actualGain = 0;			// actual gain used, per the camera
long actualTemp = 0;			// actual sensor temp, per the camera
ASI_BOOL bAuto = ASI_FALSE;		// "auto" flag returned by ASIGetControlValue, when we don't care what it is

#ifdef USE_NEW_EXPOSURE
ASI_BOOL wasAutoExposure = ASI_FALSE;

ASI_ERROR_CODE takeOneExposure(
	int cameraId,
	int exposureTimeMicroseconds,
	ASI_BOOL useAutoExposure,	// should autoexposure be used?
	unsigned char *imageBuffer, long width, long height,  // where to put image and its size
        ASI_IMG_TYPE imageType)
{
    ASI_ERROR_CODE status = (ASI_ERROR_CODE)-1;
    long bufferSize = width * height * bytesPerPixel(imageType);
    long timeout = (exposureTimeMicroseconds * 2) + 500;
    char message[1000];

    if (imageBuffer == NULL) {
        return (ASI_ERROR_CODE) -1;
    }

    sprintf(message, "setting exposure to %d microseconds, auto exposure is %d, buffer size is %ld\n", exposureTimeMicroseconds,useAutoExposure, bufferSize);
    displayDebugText(message, 1);

    setControl(cameraId, ASI_EXPOSURE, exposureTimeMicroseconds, useAutoExposure);
    status = ASIStartVideoCapture(cameraId);
    if (status == ASI_SUCCESS) {
//      printf("width = %ld, height = %ld, bufferSize = %ld\n", width, height, bufferSize);
        status = ASIGetVideoData(cameraId, imageBuffer, bufferSize, timeout);
        if (status != ASI_SUCCESS) {
            sprintf(message, "  > ERROR: Failed getting exposure data: %s\n", getRetCode(status));
            displayDebugText(message, 0);
        }
        else {
            ASIGetControlValue(cameraId, ASI_EXPOSURE, &actualExposureMicroseconds, &wasAutoExposure);
            sprintf(message, "  > got video data. Exposure value: %.0f µs (%.1f ms), actual ms=%.1lf\n", (float)actualExposureMicroseconds, (float)actualExposureMicroseconds/US_IN_MS, (double)actualExposureMicroseconds/US_IN_MS);
            //displayDebugText(message, 0);

            // The first image often says it's the correct exposure but it's not.  Not sure why.
            if (wasAutoExposure == ASI_FALSE && exposureTimeMicroseconds != actualExposureMicroseconds)
            {
                sprintf(message, "  > WARNING: not correct exposure (returned = %.1f ms, actual = %.1lf ms)\n", (float)actualExposureMicroseconds/US_IN_MS, (double)exposureTimeMicroseconds/US_IN_MS);
                displayDebugText(message, 0);
                status = (ASI_ERROR_CODE) -1;
            }
            else {
//              sprintf(message, "actual exposure value was %ld, auto exposure was %d\n", actualExposureMicroseconds, wasAutoExposure);
//              displayDebugText(message, 1);
//              // Added by linda, 9 jun 2021 to work around ASI auto exposure bug with short exposures.
//              // Auto exposures of < 30000 microseconds seem to be unreliable.
//              if (actualExposureMicroseconds != exposureTimeMicroseconds && actualExposureMicroseconds < 30000 && wasAutoExposure) {
//                  displayDebugText("retrying with short  manual exposure\n", 1);
//                  setControl(cameraId, ASI_EXPOSURE, 100, ASI_FALSE);
//                  status = ASIGetVideoData(cameraId, imageBuffer, bufferSize, 1000);
//                  ASIGetControlValue(cameraId, ASI_EXPOSURE, &actualExposureMicroseconds, &wasAutoExposure);
//              }
                ASIGetControlValue(cameraId, ASI_GAIN, &actualGain, &bAuto);
                ASIGetControlValue(cameraId, ASI_TEMPERATURE, &actualTemp, &bAuto);
            }
        }
        ASIStopVideoCapture(cameraId);
    }
    else {
      sprintf(message, "  > ERROR: Not fetching exposure data because status is %s\n", getRetCode(status));
      displayDebugText(message, 0);
    }

  return status;
}
#endif

void IntHandle(int i)
{
    bMain = false;
}

void calculateDayOrNight(const char *latitude, const char *longitude, const char *angle)
{
    char sunwaitCommand[128];
    sprintf(sunwaitCommand, "sunwait poll exit set angle %s %s %s", angle, latitude, longitude);
    dayOrNight = exec(sunwaitCommand);
    dayOrNight.erase(std::remove(dayOrNight.begin(), dayOrNight.end(), '\n'), dayOrNight.end());
}

void writeToLog(int val)
{
    std::ofstream outfile;
    outfile.open("log.txt", std::ios_base::app);
    outfile << val;
    outfile << "\n";
}

void writeTemperatureToFile(float val)
{
    std::ofstream outfile;
    outfile.open("temperature.txt", std::ios_base::trunc);
    outfile << val;
    outfile << "\n";
}

/**
 * Helper function to display debug info
**/
void displayDebugText(const char * text, int requiredLevel) {
    if (debugLevel >= requiredLevel) {

        printf("%s", text);
    }
}

//-------------------------------------------------------------------------------------------------------
//-------------------------------------------------------------------------------------------------------

int main(int argc, char *argv[])
{
    signal(SIGINT, IntHandle);
    pthread_mutex_init(&mtx_SaveImg, 0);
    pthread_cond_init(&cond_SatrtSave, 0);

    int fontname[] = { CV_FONT_HERSHEY_SIMPLEX,        CV_FONT_HERSHEY_PLAIN,         CV_FONT_HERSHEY_DUPLEX,
                       CV_FONT_HERSHEY_COMPLEX,        CV_FONT_HERSHEY_TRIPLEX,       CV_FONT_HERSHEY_COMPLEX_SMALL,
                       CV_FONT_HERSHEY_SCRIPT_SIMPLEX, CV_FONT_HERSHEY_SCRIPT_COMPLEX };
    // ECC added character representation of names for clarity.
    char const *fontnames[] = { "SIMPLEX",        "PLAIN",         "DUPEX",
                                "COMPLEX",        "TRIPLEX",       "COMPLEX_SMALL",
                                "SCRIPT_SIMPLEX", "SCRIPT_COMPLEX" };

    char buf[1024]    = { 0 };
    char bufTime[128] = { 0 };
    char bufTemp[128] = { 0 }, bufTemp2[50] = { 0 };

    // Some settings have both day and night versions, some have only one version that applies to both,
    // and some have either a day or night version but not both.
    // For settings with both versions we also keeep a "current" version that's either the day version or the
    // night version, so the code doesn't always have to check if it's day or night.
    // Day versions have "day" in the name and night versions have "night" in the name.
    // In theory, almost every setting could have day and night versions (e.g., width & height), but the chances
    // of someone wanting different versions in day and night is small.
    // #define the defaults so we can use the same value in the help message.

#define NOT_SET -1	// signifies something isn't set yet
    // All the font settings apply to both day and night.
#define DEFAULT_FONTNUMBER 0
    int fontnumber = DEFAULT_FONTNUMBER;
    int iStrLen;
#define DEFAULT_ITEXTX 15
#define DEFAULT_ITEXTY 25
    int iTextX = DEFAULT_ITEXTX;
    int iTextY = DEFAULT_ITEXTY;
#define DEFAULT_ITEXTLINEHEIGHT 30
    int iTextLineHeight = DEFAULT_ITEXTLINEHEIGHT;
    char const *ImgText   = "";
    char const *ImgExtraText   = "";
    int extraFileAge      = 0;   // 0 disables it
    char textBuffer[1024] = { 0 };
#define DEFAULT_FONTSIZE 7
    double fontsize       = DEFAULT_FONTSIZE;
#define SMALLFONTSIZE_MULTIPLIER 0.08
#define DEFAULT_LINEWIDTH 1
    int linewidth         = DEFAULT_LINEWIDTH;
#define DEFAULT_OUTLINEFONT 0
    int outlinefont       = DEFAULT_OUTLINEFONT;
    int fontcolor[3]      = { 255, 0, 0 };
    int smallFontcolor[3] = { 0, 0, 255 };
    int linetype[3]       = { CV_AA, 8, 4 };
#define DEFAULT_LINENUMBER  0
    int linenumber        = DEFAULT_LINENUMBER;

    int width                  = 0;	int originalWidth = width;
    int height                 = 0;	int originalHeight = height;
#define DEFAULT_DAYBIN           1
#define DEFAULT_NIGHTBIN         1
    int dayBin                 = DEFAULT_DAYBIN;
    int nightBin               = DEFAULT_NIGHTBIN;
    int currentBin             = NOT_SET;
#define DEFAULT_IMAGE_TYPE       1
    int Image_type             = DEFAULT_IMAGE_TYPE;
#define DEFAULT_ASIBANDWIDTH    40
    int asiBandwidth           = DEFAULT_ASIBANDWIDTH;
    int asiAutoBandwidth       = 0;		// is Auto Bandwidth on or off?
#define DEFAULT_ASIDAYEXPOSURE  (500)   // microseconds - reasonable starting point for daytime exposures
    int asiDayExposure         = DEFAULT_ASIDAYEXPOSURE;
    int asiDayAutoExposure     = 1;		// is Auto Exposure on or off for daylight?
    // There is no max day autoexposure since daylight exposures are always pretty short.
#define DEFAULT_ASINIGHTEXPOSURE (5 * US_IN_SEC)
    int asiNightExposure       = DEFAULT_ASINIGHTEXPOSURE;
#define DEFAULT_ASINIGHTMAXEXPOSURE  (10 * US_IN_MS)
    int asiNightMaxExposure    = DEFAULT_ASINIGHTMAXEXPOSURE;
    int asiNightAutoExposure   = 1;		// is Auto Exposure on or off for nighttime?
    int currentAutoExposure    = NOT_SET;	// is Auto Exposure on or off?
#define DEFAULT_ASINIGHTGAIN     150		// no need to use gain during day, so gain settings are only for night
    int asiNightGain           = DEFAULT_ASINIGHTGAIN;
#define DEFAULT_ASINIGHTMAXGAIN  200
    int asiNightMaxGain        = DEFAULT_ASINIGHTMAXGAIN;
    int asiNightAutoGain       = 0;		// is Auto Gain on or off for nighttime?
#define DEFAULT_DAYDELAY         5000
    int dayDelay               = DEFAULT_DAYDELAY; // Delay in milliseconds. Default is 5 s
#define DEFAULT_NIGHTDELAY       10000
    int nightDelay             = DEFAULT_NIGHTDELAY;   // Delay in milliseconds. Default is 10 s
    int currentDelay           = NOT_SET;
#define DEFAULT_ASIWBR           65
    int asiWBR                 = DEFAULT_ASIWBR;
#define DEFAULT_ASIWBB           85
    int asiWBB                 = DEFAULT_ASIWBB;
    int asiAutoWhiteBalance    = 0;		// is Auto White Balance on or off?
#define DEFAULT_ASIGAMMA         50		// not supported by all cameras
    int asiGamma               = DEFAULT_ASIGAMMA;

#define DEFAULT_ASIDAYBRIGHTNESS 50
    int asiDayBrightness  = DEFAULT_ASIDAYBRIGHTNESS;
#define DEFAULT_ASINIGHTBRIGHTNESS 50
    int asiNightBrightness     = DEFAULT_ASINIGHTBRIGHTNESS;
    int asiCurrentBrightness   = NOT_SET;

    int asiFlip                = 0;
    int asiCoolerEnabled       = 0;
    long asiTargetTemp         = 0;
#define DEFAULT_LATITUDE         "60.7N"
    char const *latitude       = DEFAULT_LATITUDE; //GPS Coordinates of Whitehorse, Yukon where the code was created
#define DEFAULT_LONGITUDE        "135.05W"
    char const *longitude      = DEFAULT_LONGITUDE;
#define DEFAULT_ANGLE            "-6"
    // angle of the sun with the horizon (0=sunset, -6=civil twilight, -12=nautical twilight, -18=astronomical twilight)
    char const *angle          = DEFAULT_ANGLE;
    int preview        =         0;
#define DEFAULT_SHOWTIME 1
    int showTime               = DEFAULT_SHOWTIME;
    int darkframe              = 0;		// night only
    int showDetails            = 0;
    int daytimeCapture         = 0;
    int help                   = 0;
    int quality                = NOT_SET;

    char const *bayer[]         = { "RG", "BG", "GR", "GB" };
    int CamNum                  = 0;
    int i;
    void *retval;
    bool endOfNight            = false;
    pthread_t hthdSave         = 0;

    //-------------------------------------------------------------------------------------------------------
    //-------------------------------------------------------------------------------------------------------
    setlinebuf(stdout);   // ECC added: Line buffer output so entries appear in the log immediately.
    printf("\n");
    printf("%s ******************************************\n", KGRN);
    printf("%s *** Allsky Camera Software v0.7 | 2020 ***\n", KGRN);
    printf("%s ******************************************\n\n", KGRN);
    printf("\%sCapture images of the sky with a Raspberry Pi and an ASI Camera\n", KGRN);
    printf("\n");
    printf("%sAdd -h or -help for available options\n", KYEL);
    printf("\n");
    printf("\%sAuthor: ", KNRM);
    printf("Thomas Jacquin - <jacquin.thomas@gmail.com>\n\n");
    printf("\%sContributors:\n", KNRM);
    printf("-Knut Olav Klo\n");
    printf("-Daniel Johnsen\n");
    printf("-Yang and Sam from ZWO\n");
    printf("-Robert Wagner\n");
    printf("-Michael J. Kidd - <linuxkidd@gmail.com>\n");
    printf("-Chris Kuethe\n\n");

    if (argc > 0)
    {
        for (i = 0; i < argc - 1; i++)
        {
            if (strcmp(argv[i], "-h") == 0 || strcmp(argv[i], "-help") == 0)
            {
                help = atoi(argv[i + 1]);
                i++;
            }
            else if (strcmp(argv[i], "-width") == 0)
            {
                width = atoi(argv[i + 1]);
                i++;
            }
            else if (strcmp(argv[i], "-height") == 0)
            {
                height = atoi(argv[i + 1]);
                i++;
            }
            else if (strcmp(argv[i], "-type") == 0)
            {
                Image_type = atoi(argv[i + 1]);
                i++;
            }
            else if (strcmp(argv[i], "-quality") == 0)
            {
                quality = atoi(argv[i + 1]);
                i++;
            }
            else if (strcmp(argv[i], "-nightexposure") == 0)
            {
                asiNightExposure = atoi(argv[i + 1]) * US_IN_MS;
                i++;
            }
            else if (strcmp(argv[i], "-dayexposure") == 0)
            {
                asiDayExposure = atof(argv[i + 1]) * US_IN_MS;  // ECC: allow fractions
                i++;
            }
            else if (strcmp(argv[i], "-nightmaxexposure") == 0)
            {
                asiNightMaxExposure = atoi(argv[i + 1]);
                i++;
            }
            else if (strcmp(argv[i], "-dayautoexposure") == 0)
            {
                asiDayAutoExposure = atoi(argv[i + 1]);
                i++;
            }
            else if (strcmp(argv[i], "-nightautoexposure") == 0)
            {
                asiNightAutoExposure = atoi(argv[i + 1]);
                i++;
            }
            else if (strcmp(argv[i], "-nightgain") == 0)
            {
                asiNightGain = atoi(argv[i + 1]);
                i++;
            }
            else if (strcmp(argv[i], "-nightmaxgain") == 0)
            {
                asiNightMaxGain = atoi(argv[i + 1]);
                i++;
            }
            else if (strcmp(argv[i], "-nightautogain") == 0)
            {
                asiNightAutoGain = atoi(argv[i + 1]);
                i++;
            }
            else if (strcmp(argv[i], "-gamma") == 0)
            {
                asiGamma = atoi(argv[i + 1]);
                i++;
            }
            else if (strcmp(argv[i], "-daybrightness") == 0)
            {
                asiDayBrightness = atoi(argv[i + 1]);
                i++;
            }
            else if (strcmp(argv[i], "-nightbrightness") == 0)
            {
                asiNightBrightness = atoi(argv[i + 1]);
                i++;
            }
            else if (strcmp(argv[i], "-daybin") == 0)
            {
                dayBin = atoi(argv[i + 1]);
                i++;
            }
            else if (strcmp(argv[i], "-nightbin") == 0)
            {
                nightBin = atoi(argv[i + 1]);
                i++;
            }
            else if (strcmp(argv[i], "-daydelay") == 0)
            {
                dayDelay = atoi(argv[i + 1]);
                i++;
            }
            else if (strcmp(argv[i], "-nightdelay") == 0)
            {
                nightDelay = atoi(argv[i + 1]);
                i++;
            }
            else if (strcmp(argv[i], "-wbr") == 0)
            {
                asiWBR = atoi(argv[i + 1]);
                i++;
            }
            else if (strcmp(argv[i], "-wbb") == 0)
            {
                asiWBB = atoi(argv[i + 1]);
                i++;
            }
            else if (strcmp(argv[i], "-autowhitebalance") == 0)
            {
                asiAutoWhiteBalance = atoi(argv[i + 1]);
                i++;
            }
            else if (strcmp(argv[i], "-text") == 0)
            {
                // Fix for blank text in camera settings
                if ((char)argv[i + 1][0] != '-') {
                    ImgText = (argv[i + 1]);
                    i++;
                }
            }
            else if (strcmp(argv[i], "-extratext") == 0)
            {
                if ((char)argv[i + 1][0] != '-') {
                    ImgExtraText = (argv[i + 1]);
                    i++;
                }
            }
            else if (strcmp(argv[i], "-textlineheight") == 0)
            {
                iTextLineHeight = atoi(argv[i + 1]);
                i++;
            }
            else if (strcmp(argv[i], "-extratextage") == 0)
            {
                extraFileAge = atoi(argv[i + 1]);
                i++;
            }
            else if (strcmp(argv[i], "-textx") == 0)
            {
                iTextX = atoi(argv[i + 1]);
                i++;
            }
            else if (strcmp(argv[i], "-texty") == 0)
            {
                iTextY = atoi(argv[i + 1]);
                i++;
            }
            else if (strcmp(argv[i], "-fontname") == 0)
            {
                fontnumber = atoi(argv[i + 1]);
                i++;
            }
            else if (strcmp(argv[i], "-fontcolor") == 0)
            {
                fontcolor[0] = atoi(argv[i + 1]);
                i++;
                fontcolor[1] = atoi(argv[i + 1]);
                i++;
                fontcolor[2] = atoi(argv[i + 1]);
                i++;
            }
            else if (strcmp(argv[i], "-smallfontcolor") == 0)
            {
                smallFontcolor[0] = atoi(argv[i + 1]);
                i++;
                smallFontcolor[1] = atoi(argv[i + 1]);
                i++;
                smallFontcolor[2] = atoi(argv[i + 1]);
                i++;
            }
            else if (strcmp(argv[i], "-fonttype") == 0)
            {
                linenumber = atoi(argv[i + 1]);
                i++;
            }
            else if (strcmp(argv[i], "-fontsize") == 0)
            {
                fontsize = atof(argv[i + 1]);
                i++;
            }
            else if (strcmp(argv[i], "-fontline") == 0)
            {
                linewidth = atoi(argv[i + 1]);
                i++;
            }
            else if (strcmp(argv[i], "-outlinefont") == 0)
            {
                outlinefont = atoi(argv[i + 1]);
		if (outlinefont != 0)
		    outlinefont = 1;
                i++;
            }
            else if (strcmp(argv[i], "-flip") == 0)
            {
                asiFlip = atoi(argv[i + 1]);
                i++;
            }
            else if (strcmp(argv[i], "-usb") == 0)
            {
                asiBandwidth = atoi(argv[i + 1]);
                i++;
            }
            else if (strcmp(argv[i], "-autousb") == 0)
            {
                asiAutoBandwidth = atoi(argv[i + 1]);
                i++;
            }
            else if (strcmp(argv[i], "-filename") == 0)
            {
                fileName = (argv[i + 1]);
                i++;
            }
            else if (strcmp(argv[i], "-latitude") == 0)
            {
                latitude = argv[i + 1];
                i++;
            }
            else if (strcmp(argv[i], "-longitude") == 0)
            {
                longitude = argv[i + 1];
                i++;
            }
            else if (strcmp(argv[i], "-angle") == 0)
            {
                angle = argv[i + 1];
                i++;
            }
            else if (strcmp(argv[i], "-preview") == 0)
            {
                preview = atoi(argv[i + 1]);
                i++;
            }
            else if (strcmp(argv[i], "-debuglevel") == 0)
            {
                debugLevel = atoi(argv[i + 1]);
                i++;
            }
            else if (strcmp(argv[i], "-time") == 0)
            {
                showTime = atoi(argv[i + 1]);
                i++;
            }
            else if (strcmp(argv[i], "-timeformat") == 0)
            {
                // May contain multiple arguments separated by "_" which means a space.
                strcpy(timeFormat, argv[i + 1]);
                i++;
                int z = 0;
                while (timeFormat[z] != '\0') {
                    if (timeFormat[z] == '_') timeFormat[z] = ' ';
                    z++;
                }
            }
            else if (strcmp(argv[i], "-darkframe") == 0)
            {
                darkframe = atoi(argv[i + 1]);
                i++;
            }
            else if (strcmp(argv[i], "-showDetails") == 0)
            {
                showDetails = atoi(argv[i + 1]);
                i++;
            }
            else if (strcmp(argv[i], "-daytime") == 0)
            {
                daytimeCapture = atoi(argv[i + 1]);
                i++;
            }
	    else if (strcmp(argv[i], "-coolerEnabled") == 0)
            {
                asiCoolerEnabled = atoi(argv[i + 1]);
                i++;
            }
	    else if (strcmp(argv[i], "-targetTemp") == 0)
            {
                asiTargetTemp = atol(argv[i + 1]);
                i++;
            }
        }
    }

    if (help == 1)
    {
        printf("%sAvailable Arguments:\n", KYEL);
        printf(" -width                             - Default = 0 = Camera Max Width\n");
        printf(" -height                            - Default = 0 = Camera Max Height\n");
        printf(" -dayexposure                       - Default = %d - Time in µs (equals to %.4f sec)\n", DEFAULT_ASIDAYEXPOSURE, (float)DEFAULT_ASIDAYEXPOSURE/US_IN_SEC);
        printf(" -nightexposure                     - Default = %d - Time in µs (equals to %.4f sec)\n", DEFAULT_ASINIGHTEXPOSURE, (float)DEFAULT_ASINIGHTEXPOSURE/US_IN_SEC);
        printf(" -nightmaxexposure                  - Default = %d - Time in ms (equals to %.1f sec)\n", DEFAULT_ASINIGHTMAXEXPOSURE, (float)DEFAULT_ASINIGHTMAXEXPOSURE/US_IN_MS);

        printf(" -dayautoexposure                   - Default = 1 - Set to 1 to enable daytime auto Exposure\n");
        printf(" -nightautoexposure                 - Default = 1 - Set to 1 to enable nighttime auto Exposure\n");
        printf(" -nightgain                         - Default = %d\n", DEFAULT_ASINIGHTGAIN);
        printf(" -nightmaxgain                      - Default = %d\n", DEFAULT_ASINIGHTMAXGAIN);
        printf(" -nightautogain                     - Default = 0 - Set to 1 to enable auto Gain\n");
        printf(" -coolerEnabled                     - Set to 1 to enable cooler (works on cooled cameras only)\n");
        printf(" -targetTemp                        - Target temperature in degrees C (works on cooled cameras only)\n");
        printf(" -gamma                             - Default = %d\n", DEFAULT_ASIGAMMA);
        printf(" -daybrightness                     - Default = %d (range: 0 - 600)\n", DEFAULT_ASIDAYBRIGHTNESS);
        printf(" -nightbrightness                   - Default = %d (range: 0 - 600)\n", DEFAULT_ASINIGHTBRIGHTNESS);
        printf(" -wbr                               - Default = %d   - manual White Balance Red\n", DEFAULT_ASIWBR);
        printf(" -wbb                               - Default = %d   - manual White Balance Blue\n", DEFAULT_ASIWBB);
        printf(" -autowhitebalance                  - Default = 0 - Set to 1 to enable auto White Balance\n");
        printf(" -daybin                            - Default = %d    - 1 = binning OFF (1x1), 2 = 2x2 binning, 4 = 4x4 binning\n", DEFAULT_DAYBIN);
        printf(" -nightbin                          - Default = %d    - 1 = binning OFF (1x1), 2 = 2x2 binning, 4 = 4x4 binning\n", DEFAULT_NIGHTBIN);
        printf(" -dayDelay                          - Default = %d   - Delay between daytime images in milliseconds - 5000 = 5 sec.\n", DEFAULT_DAYDELAY);
        printf(" -nightDelay                        - Default = %d   - Delay between night images in milliseconds - 1000 = 1 sec.\n", DEFAULT_NIGHTDELAY);
        printf(" -type = Image Type                 - Default = %d    - 0 = RAW8,  1 = RGB24,  2 = RAW16\n", DEFAULT_IMAGE_TYPE);
        printf(" -quality                           - Default PNG=3, JPG=95, Values: PNG=0-9, JPG=0-100\n");
        printf(" -usb = USB Speed                   - Default = %d   - Values between 40-100, This is "
               "BandwidthOverload\n", DEFAULT_ASIBANDWIDTH);
        printf(" -autousb                           - Default = 0 - Set to 1 to enable auto USB Speed\n");
        printf(" -filename                          - Default = %s\n", DEFAULT_FILENAME);
        printf(" -flip                              - Default = 0    - 0 = Orig, 1 = Horiz, 2 = Verti, 3 = Both\n");
        printf("\n");
        printf(" -text                              - Default =      - Character/Text Overlay. Use Quotes.  Ex. -c \"Text Overlay\"\n");
        printf(" -extratext                         - Default =      - Full Path to extra text to display\n");
        printf(" -extratextage                      - Default = n/a  - If the extra file is not updated after this many seconds its contents will not be displayed. Set to 0 to disable\n");
        printf(" -textlineheight                    - Default = %d   - Text Line Height in Pixels\n", DEFAULT_ITEXTLINEHEIGHT);
        printf(" -textx                             - Default = %d   - Text Placement Horizontal from LEFT in Pixels\n", DEFAULT_ITEXTX);
        printf(" -texty = Text Y                    - Default = %d   - Text Placement Vertical from TOP in Pixels\n", DEFAULT_ITEXTY);
        printf(" -fontname = Font Name              - Default = %d   - Font Types (0-7), Ex. 0 = simplex, 4 = triplex, 7 = script\n", DEFAULT_FONTNUMBER);
        printf(" -fontcolor = Font Color            - Default = 255 0 0  - Text blue (BGR)\n");
        printf(" -smallfontcolor = Small Font Color - Default = 0 0 255  - Text red (BGR)\n");
        printf(" -fonttype = Font Type              - Default = %d    - Font Line Type,(0-2), 0 = AA, 1 = 8, 2 = 4\n", DEFAULT_LINENUMBER);
        printf(" -fontsize                          - Default = %d    - Text Font Size\n", DEFAULT_FONTSIZE);
        printf(" -fontline                          - Default = %d    - Text Font Line Thickness\n", DEFAULT_LINEWIDTH);
        printf(" -outlinefont                       - Default = %d    - TSet to 1 to enable outline font\n", DEFAULT_OUTLINEFONT);
        //printf(" -bgc = BG Color                    - Default =      - Text Background Color in Hex. 00ff00 = Green\n");
        //printf(" -bga = BG Alpha                    - Default =      - Text Background Color Alpha/Transparency 0-100\n");
        printf("\n");
        printf("\n");
        printf(" -latitude                          - Default = %7s (Whitehorse) - Latitude of the camera.\n", DEFAULT_LATITUDE);
        printf(" -longitude                         - Default = %7s (Whitehorse) - Longitude of the camera\n", DEFAULT_LONGITUDE);
        printf(" -angle                             - Default = %s - Angle of the sun below the horizon.\n", DEFAULT_ANGLE);
        printf("   -6=civil twilight\n   -12=nautical twilight\n   -18=astronomical twilight\n");
        printf("\n");
        printf(" -preview                           - Set to 1 to preview the captured images. Only works with a Desktop Environment\n");
        printf(" -time                              - Set to 1 to add the time to the image. Combine with Text X and Text Y for placement\n");
        printf(" -timeformat                        - Format the optional time is displayed in; default is '%s'\n", DEFAULT_TIMEFORMAT);
        printf(" -darkframe                         - Set to 1 to disable time and text overlay and take dark frames instead.\n");
        printf(" -showDetails                       - Set to 1 to display the metadata on the image (time, sensor temp, etc.)\n");
        printf(" -debuglevel                        - Default = 0. Set to 1,2 or 3 for debug level\n");
        printf("%sUsage:\n", KRED);
        printf(" ./capture -width 640 -height 480 -exposure 5000000 -gamma 50 -type 1 -bin 1 -filename Lake-Laberge.PNG\n\n");
    }
    printf("%s", KNRM);

    const char *ext = strrchr(fileName, '.');
    if (strcmp(ext + 1, "jpg") == 0 || strcmp(ext + 1, "JPG") == 0 || strcmp(ext + 1, "jpeg") == 0 ||
        strcmp(ext + 1, "JPEG") == 0)
    {
        compression_parameters.push_back(CV_IMWRITE_JPEG_QUALITY);
        if (quality == NOT_SET)
        {
            quality = 95;
        }
    }
    else
    {
        compression_parameters.push_back(CV_IMWRITE_PNG_COMPRESSION);
        if (quality == NOT_SET)
        {
            quality = 3;
        }
    }
    compression_parameters.push_back(quality);

    int numDevices = ASIGetNumOfConnectedCameras();
    if (numDevices <= 0)
    {
        printf("\nERROR: No Connected Camera...\n");
        return 1;   // ECC added - If there are no cameras we can't do anything.
    }

    printf("\nListing Attached Cameras:\n");

    ASI_CAMERA_INFO ASICameraInfo;

    for (i = 0; i < numDevices; i++)
    {
        ASIGetCameraProperty(&ASICameraInfo, i);
        printf("- %d %s\n", i, ASICameraInfo.Name);
    }

    if (ASIOpenCamera(CamNum) != ASI_SUCCESS)
    {
        printf("Open Camera ERROR, Check that you have root permissions!\n");
        return -1;      // ECC added - can't do anything so might as well exit.
    }

    printf("\n%s Information:\n", ASICameraInfo.Name);
    int iMaxWidth, iMaxHeight;
    double pixelSize;
    iMaxWidth  = ASICameraInfo.MaxWidth;
    iMaxHeight = ASICameraInfo.MaxHeight;
    pixelSize  = ASICameraInfo.PixelSize;
    printf("- Resolution:%dx%d\n", iMaxWidth, iMaxHeight);
    printf("- Pixel Size: %1.1fμm\n", pixelSize);
    printf("- Supported Bin: ");
    for (int i = 0; i < 16; ++i)
    {
        if (ASICameraInfo.SupportedBins[i] == 0)
        {
            break;
        }
        printf("%d ", ASICameraInfo.SupportedBins[i]);
    }
    printf("\n");

    if (ASICameraInfo.IsColorCam)
    {
        printf("- Color Camera: bayer pattern:%s\n", bayer[ASICameraInfo.BayerPattern]);
    }
    else
    {
        printf("- Mono camera\n");
    }
    if (ASICameraInfo.IsCoolerCam)
    {
        printf("- Camera with cooling capabilities\n");
    }

    const char *ver = ASIGetSDKVersion();
    printf("- SDK version %s\n", ver);

    if (ASIInitCamera(CamNum) == ASI_SUCCESS)
    {
        printf("- Initialise Camera OK\n");
    }
    else
    {
        printf("- Initialise Camera ERROR\n");
        return -1;      // ECC added - can't do anything so might as well exit.
    }

    ASIGetNumOfControls(CamNum, &iNumOfCtrl);
    if (0)	// this is really only needed for debugging
    {
    	// ECC: Display the control capabilities, not just the names, for debugging.
        printf("Control Caps:\n");
        for (i = 0; i < iNumOfCtrl; i++)
        {
            ASIGetControlCaps(CamNum, i, &ControlCaps);
            printf("- %s:\n", ControlCaps.Name);
            printf("   - MinValue = %ld\n", ControlCaps.MinValue);
            printf("   - MaxValue = %ld\n", ControlCaps.MaxValue);
            printf("   - DefaultValue = %ld\n", ControlCaps.DefaultValue);
            printf("   - IsAutoSupported = %d\n", ControlCaps.IsAutoSupported);
            printf("   - IsWritable = %d\n", ControlCaps.IsWritable);
            printf("   - ControlType = %d\n", ControlCaps.ControlType);
        }
    }

    if (width == 0 || height == 0)
    {
        width  = iMaxWidth;	originalWidth = width;
        height = iMaxHeight;	originalHeight = height;
    }

    ASIGetControlValue(CamNum, ASI_TEMPERATURE, &actualTemp, &bAuto);
    printf("- Sensor temperature:%0.2f\n", (float)actualTemp / 10.0);

    const char *sType;
    if (Image_type == ASI_IMG_RAW16)
    {
        sType = "ASI_IMG_RAW16";
    }
    else if (Image_type == ASI_IMG_RGB24)
    {
        sType = "ASI_IMG_RGB24";
    }
    else
    {
        sType = "ASI_IMG_RAW8";
    }

    //-------------------------------------------------------------------------------------------------------
    //-------------------------------------------------------------------------------------------------------

    printf("%s", KGRN);
    printf("\nCapture Settings:\n");
    printf(" Image Type: %s\n", sType);
    printf(" Resolution (before any binning): %dx%d\n", width, height);
    printf(" Quality: %d\n", quality);
    printf(" Exposure (day): %1.3fms\n", (float)asiDayExposure / US_IN_MS);
    printf(" Auto Exposure (day): %d\n", asiDayAutoExposure);
    printf(" Exposure (night): %1.0fms\n", round(asiNightExposure / US_IN_MS));
    printf(" Max Exposure (night): %dms\n", asiNightMaxExposure);
    printf(" Auto Exposure (night): %d\n", asiNightAutoExposure);
    printf(" Gain (night only): %d\n", asiNightGain);
    printf(" Auto Gain (night only): %d\n", asiNightAutoGain);
    printf(" Max Gain (night only): %d\n", asiNightMaxGain);
    printf(" Cooler Enabled: %d\n", asiCoolerEnabled);
    printf(" Target Temperature: %ldC\n", asiTargetTemp);
    printf(" Brightness (day): %d\n", asiDayBrightness);
    printf(" Brightness (night): %d\n", asiNightBrightness);
    printf(" Gamma: %d\n", asiGamma);
    printf(" WB Red: %d\n", asiWBR);
    printf(" WB Blue: %d\n", asiWBB);
    printf(" Auto WB: %d\n", asiAutoWhiteBalance);
    printf(" Binning (day): %d\n", dayBin);
    printf(" Binning (night): %d\n", nightBin);
    printf(" Delay (day): %dms\n", dayDelay);
    printf(" Delay (night): %dms\n", nightDelay);
    printf(" USB Speed: %d\n", asiBandwidth);
    printf(" Auto USB Speed: %d\n", asiAutoBandwidth);
    printf(" Text Overlay: %s\n", ImgText);
    printf(" Text Extra Filename: %s\n", ImgExtraText);
    printf(" Text Extra Filename Age: %d\n", extraFileAge);
    printf(" Text Line Height %dpx\n", iTextLineHeight);
    printf(" Text Position: %dpx left, %dpx top\n", iTextX, iTextY);
    printf(" Font Name:  %d (%s)\n", fontname[fontnumber], fontnames[fontnumber]);
    printf(" Font Color: %d , %d, %d\n", fontcolor[0], fontcolor[1], fontcolor[2]);
    printf(" Small Font Color: %d , %d, %d\n", smallFontcolor[0], smallFontcolor[1], smallFontcolor[2]);
    printf(" Font Line Type: %d\n", linetype[linenumber]);
    printf(" Font Size: %1.1f\n", fontsize);
    printf(" Font Line: %d\n", linewidth);
    printf(" Outline Font : %d\n", outlinefont);
    printf(" Flip Image: %d\n", asiFlip);
    printf(" Filename: %s\n", fileName);
    printf(" Latitude: %s\n", latitude);
    printf(" Longitude: %s\n", longitude);
    printf(" Sun Elevation: %s\n", angle);
    printf(" Preview: %d\n", preview);
    printf(" Time: %d\n", showTime);
    printf(" Time Format: %s\n", timeFormat);
    printf(" Darkframe: %d\n", darkframe);
    printf(" Show Details: %d\n", showDetails);
    printf(" Debug Level: %d\n", debugLevel);
    printf("%s", KNRM);

    //-------------------------------------------------------------------------------------------------------
    //-------------------------------------------------------------------------------------------------------
    // ECC: only set controls supported by this camera, check for errors, and allow more auto settings.
    // Other calls to setControl() are done after we know if we're in daytime or nighttime.
    setControl(CamNum, ASI_BANDWIDTHOVERLOAD, asiBandwidth, asiAutoBandwidth == 1 ? ASI_TRUE : ASI_FALSE);
    setControl(CamNum, ASI_HIGH_SPEED_MODE, 0, ASI_FALSE);  // ECC added to try and get rid of lines on some pictures
    setControl(CamNum, ASI_WB_R, asiWBR, asiAutoWhiteBalance == 1 ? ASI_TRUE : ASI_FALSE);
    setControl(CamNum, ASI_WB_B, asiWBB, asiAutoWhiteBalance == 1 ? ASI_TRUE : ASI_FALSE);
    setControl(CamNum, ASI_GAMMA, asiGamma, ASI_FALSE);
    setControl(CamNum, ASI_FLIP, asiFlip, ASI_FALSE);

    if (ASICameraInfo.IsCoolerCam)
    {
        ASI_ERROR_CODE err = setControl(CamNum, ASI_COOLER_ON, asiCoolerEnabled == 1 ? ASI_TRUE : ASI_FALSE, ASI_FALSE);
        if (err != ASI_SUCCESS)
        {
            printf("%s", KRED);
            printf(" Could not enable cooler\n");
            printf("%s", KNRM);
        }
        err = setControl(CamNum, ASI_TARGET_TEMP, asiTargetTemp, ASI_FALSE);
        if (err != ASI_SUCCESS)
        {
            printf("%s", KRED);
            printf(" Could not set cooler temperature\n");
            printf("%s", KNRM);
        }
    }

    pthread_t thread_display = 0;
    if (preview == 1)
    {
        bDisplay = 1;
        pthread_create(&thread_display, NULL, Display, (void *)&pRgb);
    }

    if (!bSaveRun)
    {
        bSaveRun = true;
        if (pthread_create(&hthdSave, 0, SaveImgThd, 0) != 0)
        {
            bSaveRun = false;
        }
    }

    // Initialization
    int currentExposure = NOT_SET;	// the last exposure, will be next one
    int exitCode        = 0;    // Exit code for main()
    int numErrors       = 0;    // ECC: number of errors in a row.
    int maxErrors       = 2;    // ECC: max number of errors in a row before we exit
    int originalITextX = iTextX;
    int originalITextY = iTextY;
    int originalFontsize = fontsize;
    int originalLinewidth = linewidth;

    while (bMain)
    {
        bool needCapture = true;
        std::string lastDayOrNight;
        int captureTimeout = -1;

        // Find out if it is currently DAY or NIGHT
        calculateDayOrNight(latitude, longitude, angle);

        lastDayOrNight = dayOrNight;
        printf("\n");
        if (dayOrNight == "DAY")
        {
            // Setup the daytime capture parameters
            if (endOfNight == true)
            {
                sprintf(textBuffer, "Processing end of night data\n");   // ECC added
                displayDebugText(textBuffer, 0);
                system("scripts/endOfNight.sh &");
                endOfNight = false;
            }
            if (daytimeCapture != 1)
            {
                needCapture = false;
                sprintf(textBuffer, "It's daytime... we're not saving images\n");
                displayDebugText(textBuffer, 0);
                usleep(dayDelay * US_IN_MS);
            }
            else
            {
                sprintf(textBuffer, "Starting daytime capture\nSaving images with delay of %d ms\n\n", dayDelay);
                displayDebugText(textBuffer, 0);
                currentAutoExposure = asiDayAutoExposure;
                asiCurrentBrightness = asiDayBrightness;
                currentDelay = dayDelay;
                currentBin = dayBin;

                // If we went from Night to Day, then currentExposure will be the last exposure so leave it,
                // so there's a seamless change from Night to Day, i.e., if the exposure was fine a minute ago
                // it will likely be fine now.
                // On the other hand, if this program just started use the default or what the user specified.
                if (currentExposure == NOT_SET)	// no prior exposures
                    currentExposure = asiDayExposure;
                    // Per ASICamera2.h:    -1 means wait forever. this value is recommend set to (exposure*2) + 500ms
                captureTimeout = (currentExposure * 2) + 500;
#ifndef USE_HISTOGRAM
                setControl(CamNum, ASI_EXPOSURE, currentExposure, asiDayAutoExposure == 1 ? ASI_TRUE : ASI_FALSE);
#endif
                setControl(CamNum, ASI_AUTO_MAX_EXP, 60 * US_IN_MS, ASI_FALSE);    // reset in case set by nightiime
                setControl(CamNum, ASI_GAIN, 0, ASI_FALSE);	// we don't use gain during the day
            }
        }
        else if (dayOrNight == "NIGHT")
        {
            // Setup the night time capture parameters
            if (asiNightAutoExposure == 1)
            {
                printf("Saving auto exposed night images with delay of %d ms\n\n", nightDelay);
            }
            else
            {
                printf("Saving %ds exposure night images with delay of %d ms\n\n", (int)round(currentExposure / US_IN_SEC), nightDelay);
            }
            // Set exposure value for night time capture
            currentAutoExposure = asiNightAutoExposure;
            asiCurrentBrightness = asiNightBrightness;
            currentDelay = nightDelay;
            currentBin = nightBin;
            if (currentExposure == NOT_SET)	// no prior exposures
             	currentExposure = asiNightExposure;
#ifndef USE_HISTOGRAM
            setControl(CamNum, ASI_EXPOSURE, currentExposure, asiNightAutoExposure == 1 ? ASI_TRUE : ASI_FALSE);
#endif
            currentAutoExposure = asiNightAutoExposure;	// are we in auto-exposure mode?
            setControl(CamNum, ASI_GAIN, asiNightGain, asiNightAutoGain == 1 ? ASI_TRUE : ASI_FALSE);
        }
        setControl(CamNum, ASI_BRIGHTNESS, asiCurrentBrightness, ASI_FALSE);

        // Adjusting variables for chosen binning
        height    = originalHeight / currentBin;
        width     = originalWidth / currentBin;
        iTextX    = originalITextX / currentBin;
        iTextY    = originalITextY / currentBin;
        fontsize  = originalFontsize / currentBin;
        linewidth = originalLinewidth / currentBin;

        if (Image_type == ASI_IMG_RAW16)
        {
            pRgb.create(cvSize(width, height), CV_16UC1);
        }
        else if (Image_type == ASI_IMG_RGB24)
        {
            pRgb.create(cvSize(width, height), CV_8UC3);
        }
        else
        {
            pRgb.create(cvSize(width, height), CV_8UC1);
        }
        if (Image_type != ASI_IMG_RGB24 && Image_type != ASI_IMG_RAW16)
        {
            // xxxxxxxx "buf" hasn't been set to anything yet, so strlen() is 0,
            // and nothing ever sets "buf".
            iStrLen     = strlen(buf);
            CvRect rect = cvRect(iTextX, iTextY - 15, iStrLen * 11, 20);
            cv::Mat roi = pRgb(rect);
            roi.setTo(cv::Scalar(180, 180, 180));
        }
        ASISetROIFormat(CamNum, width, height, currentBin, (ASI_IMG_TYPE)Image_type);

        printf("Press Ctrl+C to stop\n\n");

        if (needCapture)
        {
            ASI_ERROR_CODE retCode;    // ECC added
#ifndef USE_NEW_EXPOSURE
            retCode = ASIStartVideoCapture(CamNum);
            if (retCode != ASI_SUCCESS) // ECC added check
            {
                sprintf(textBuffer,"----- ASIStartVideoCapture(%d) failed, retCode=%s\n", CamNum, getRetCode(retCode));
                displayDebugText(textBuffer, 0);
                bMain = 0;
                exitCode = 1;
                break;  // get out of bMain loop
            }
#endif

            // ECC: Here and below, indent sub-messages with "  > " so it's clear they go with the un-indented line.
            // This simply makes it easier to see things in the log file.

            // ECC: As of April 2021 there's a bug that causes the first 3 images to be identical,
            // so take 3 short ones but don't save them.
            // On the ASI178MC the shortest time is 10010 µs; it may be higher on other cameras,
            // so use a higher value like 30,000 µs to be safe.
#define SHORT_EXPOSURE 30000
            printf("===Taking 3 images to clear buffer...\n");
#ifndef USE_NEW_EXPOSURE
            setControl(CamNum, ASI_EXPOSURE, SHORT_EXPOSURE, ASI_FALSE);
#endif

//            FILE *fp = fopen("exposures.csv", "w");
//            long histogram[256];
//            int peak;
//            if (fp != NULL) {
//              for (int i = 100; i < 500000; i += 50) {
//                takeOneExposure(CamNum, i, ASI_FALSE, pRgb.data, width, height, (ASI_IMG_TYPE) Image_type);
//                computeHistogram(pRgb.data, width, height, (ASI_IMG_TYPE) Image_type, histogram);
//                peak = calculateHistogramPeak(histogram);
//                fprintf(fp, "%d, %d\n", i, peak);
//                if (i % 1000 == 0) {
//                  printf("%d microseconds\n", i);
//                  fflush(fp);
//                }
//              }
//              fclose(fp);
//            }
//            else {
//              printf("couldn't open file!!!!!!!\n");
//            }
//  
//            exit(0);
       
            for (i=1; i <= 3; i++)
            {
                if (i > 1) sleep(1);    // Sometimes the 2nd or 3rd image fail, so see if sleeping first helps.
#ifdef USE_NEW_EXPOSURE
                retCode = takeOneExposure(CamNum, SHORT_EXPOSURE, ASI_FALSE, pRgb.data, width, height, (ASI_IMG_TYPE) Image_type);
#else
                retCode = ASIGetVideoData(CamNum, pRgb.data, pRgb.step[0] * pRgb.rows, 10000);
#endif
                if (retCode != ASI_SUCCESS)
                {
#ifndef USE_NEW_EXPOSURE	// new takeOneExposure() already printed error msg
                    printf("  > ASIGetVideoData() failed on image %d, retCode=%s\n", i, getRetCode(retCode));
#endif
                    numErrors++;
                    sleep(2);
                }
            }
            if (numErrors >= maxErrors)
            {
                bMain = 0;
                exitCode = 2;
                break;
            }

            // Restore correct exposure times.
            setControl(CamNum, ASI_EXPOSURE, currentExposure, currentAutoExposure == 1 ? ASI_TRUE : ASI_FALSE);
            printf("...DONE.  Resetting exposure to %d µs\n", currentExposure);
            // END of bug code

            while (bMain && lastDayOrNight == dayOrNight)
            {
                sprintf(textBuffer, "STARTING EXPOSURE at: %s\n", getTime2());   // ECC added
                displayDebugText(textBuffer, 0);
#ifdef USE_NEW_EXPOSURE
                retCode = takeOneExposure(CamNum, currentExposure, ASI_FALSE, pRgb.data, width, height, (ASI_IMG_TYPE) Image_type);
#else
                retCode = ASIGetVideoData(CamNum, pRgb.data, pRgb.step[0] * pRgb.rows, captureTimeout);
#endif
                if (retCode == ASI_SUCCESS)
                {
                    numErrors = 0;

#ifndef USE_NEW_EXPOSURE
                    // Read current camera parameters
                    ASIGetControlValue(CamNum, ASI_EXPOSURE, &actualExposureMicroseconds, &bAuto);
                    ASIGetControlValue(CamNum, ASI_GAIN, &actualGain, &bAuto);
                    ASIGetControlValue(CamNum, ASI_TEMPERATURE, &actualTemp, &bAuto);
#endif

#ifdef USE_HISTOGRAM
                    long histogram[256];
                    computeHistogram(pRgb.data, width, height, (ASI_IMG_TYPE) Image_type, histogram);
                    int peak = calculateHistogramPeak(histogram);
                    //int prevPeak = 255;    // previous peak - to detect unusually large jumps

                    int attempts = 0;
                    int newExposure = 0;
                    //int prevExposure = currentExposure;
                    const int minExposure = 100;
                    const int minAcceptableHistogram = 120;
//                    const int desiredHistogram = 127;
                    const int maxAcceptableHistogram = 132;
                    int tempMinExposure = minExposure;
                    int tempMaxExposure = asiNightMaxExposure *  US_IN_MS;

                    while (peak < minAcceptableHistogram || peak > maxAcceptableHistogram) {
                        if (peak == 255) {
                          newExposure = currentExposure * 0.75;
                          tempMaxExposure = currentExposure - 10;
                        }
                        else if (peak < 10) {
                          newExposure = currentExposure * 2;
                          tempMinExposure = currentExposure + 10;
                        }
                        else if (peak < minAcceptableHistogram) {
                            newExposure = currentExposure * 1.1;
                            tempMinExposure = currentExposure + 10;
                        }
                        else if (peak > maxAcceptableHistogram) {
                          newExposure = currentExposure * 0.9;
                          tempMaxExposure = currentExposure - 10;
                        }

			newExposure = round10(newExposure);
                        newExposure = std::max(tempMinExposure, newExposure);
                        newExposure = std::min(newExposure, tempMaxExposure);

			sprintf(textBuffer, "  > attempt %i,    current exposure %d µs,   peak %d,    newExposure %d µs,   temp min exposure %d µs,   tempMaxExposure %d µs\n", attempts, currentExposure, peak, newExposure, tempMinExposure, tempMaxExposure);
                        displayDebugText(textBuffer, 1);

			if (newExposure == currentExposure)
			  break;  // we can't find a better exposure so stick with this one

// xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
// xxxxxxxxxxx are minExposure and maxExposure both in same units, e.g., microseconds?
// xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

                        if (++attempts > 9 || ((newExposure == currentExposure) && (currentExposure == minExposure || currentExposure == (asiNightMaxExposure * US_IN_MS))))
                            break;

                        currentExposure = newExposure;

                        sprintf(textBuffer, "  > !!! Retrying with manual exposure of %d microseconds\n", currentExposure);
                        displayDebugText(textBuffer, 1);
                        takeOneExposure(CamNum, currentExposure, ASI_FALSE, pRgb.data, width, height, (ASI_IMG_TYPE) Image_type);

                        computeHistogram(pRgb.data, width, height, (ASI_IMG_TYPE) Image_type, histogram);
                        peak = calculateHistogramPeak(histogram);

//                        if (peak > (prevPeak * 1.25))    // unexpected large jump in peak - ignore it
//                        {
// xxxxxxxxxxxxxxx not sure what to do - possibly have next exposure be slightly less than this one?
                            // ?? currentExposure = prevExposure * .90;   // lower little
//                            printf("  > >>> New peak had large jump (from %d to %d) - ignoring; setting currentExposure to %d\n", prevPeak, peak, currentExposure);
                            // ?? peak = prevPeak;
//                        } else {
//                            prevPeak = peak;
//                            prevExposure = currentExposure;
//                        }

                    }
#endif

#ifndef USE_NEW_EXPOSURE
                    // finish reading camera parameters
                    ASIGetControlValue(cameraId, ASI_EXPOSURE, &actualExposureMicroseconds, &wasAutoExposure);
                    ASIGetControlValue(CamNum, ASI_GAIN, &actualGain, &bAuto);
                    ASIGetControlValue(CamNum, ASI_TEMPERATURE, &actualTemp, &bAuto);
#endif

                    // Write temperature to file
                    writeTemperatureToFile((float)actualTemp / 10.0);

                    // Get Current Time for overlay
                    sprintf(bufTime, "%s", getTime2());     // ECC: added "2"

                    if (darkframe != 1)
                    {
                        // If darkframe mode is off, add overlay text to the image
                        int iYOffset = 0;
                        // ECC: Why are these commented out?
                        // cvText(pRgb, ImgText, iTextX, iTextY+(iYOffset/currentBin), fontsize, linewidth, linetype[linenumber], fontname[fontnumber], fontcolor, Image_type);
                        // iYOffset+=iTextLineHeight;
                        if (showTime == 1)
                        {
                            cvText(pRgb, bufTime, iTextX, iTextY + (iYOffset / currentBin), fontsize * 0.1, linewidth,
                                   linetype[linenumber], fontname[fontnumber], fontcolor, Image_type, outlinefont);
                            iYOffset += iTextLineHeight;
                        }

                        if (showDetails == 1)
                        {
                            sprintf(bufTemp, "Sensor %.1fC (%.1fF)", (float)actualTemp / 10, (((float)actualTemp / 10) * 1.8) + 32); // ECC added degrees F
                            cvText(pRgb, bufTemp, iTextX, iTextY + (iYOffset / currentBin), fontsize * SMALLFONTSIZE_MULTIPLIER, linewidth,
                                   linetype[linenumber], fontname[fontnumber], smallFontcolor, Image_type, outlinefont);
                            iYOffset += iTextLineHeight;

                            // ECC: for debugging, indicate when the time to take the exposure is less than the reported exposure time
                            if (actualExposureMicroseconds == currentExposure)
                                bufTemp2[0] = '\0';
                            else
                                sprintf(bufTemp2, " actual %.1lf ms)", (double)actualExposureMicroseconds / US_IN_MS);
                            if (actualExposureMicroseconds >= (1 * US_IN_SEC))  // ECC: display in seconds if >= 1 second, else in ms
                                sprintf(bufTemp, "Exposure %.2f s%s", (float)currentExposure / US_IN_SEC, bufTemp2);
                            else
                                sprintf(bufTemp, "Exposure %.2f ms%s", (float)currentExposure / US_IN_MS, bufTemp2);
                            // ECC added - indicate if in auto exposure mode
                            if (currentAutoExposure == 1) strcat(bufTemp, " (auto)");
                            cvText(pRgb, bufTemp, iTextX, iTextY + (iYOffset / currentBin), fontsize * SMALLFONTSIZE_MULTIPLIER, linewidth,
                                   linetype[linenumber], fontname[fontnumber], smallFontcolor, Image_type, outlinefont);
                            iYOffset += iTextLineHeight;

                            sprintf(bufTemp, "Gain %d", (int)actualGain);
                            // ECC added - indicate if in auto gain mode
                            if (asiNightAutoGain == 1) strcat(bufTemp, " (auto)");
                            cvText(pRgb, bufTemp, iTextX, iTextY + (iYOffset / currentBin), fontsize * SMALLFONTSIZE_MULTIPLIER, linewidth,
                                   linetype[linenumber], fontname[fontnumber], smallFontcolor, Image_type, outlinefont);
                            iYOffset += iTextLineHeight;

#ifdef USE_HISTOGRAM
                            sprintf(bufTemp, "Histogram peak: %d", peak);
                            cvText(pRgb, bufTemp, iTextX, iTextY + (iYOffset / currentBin), fontsize * SMALLFONTSIZE_MULTIPLIER, linewidth,
                                   linetype[linenumber], fontname[fontnumber], smallFontcolor, Image_type, outlinefont);
                            iYOffset += iTextLineHeight;
#endif
                        }
                        /**
                         * Display extra text if required. The extra text is read from the provided file. If the
                         * age of the file exceeds the specified limit then the text in the file is not displayed
                         * this is to prevent situations where the code updating the text file stops working.
                         **/
                        if (ImgExtraText[0] != '\0') {
                            bool bUseExtraFile = true;
                            if (access(ImgExtraText, F_OK ) == -1 ) {
                                bUseExtraFile = false;
                                 displayDebugText("Extra Text File Does Not Exist So Ignoring It\n", 1);
                            } else {
                                if (access(ImgExtraText, R_OK ) == -1 ) {
                                    displayDebugText("Cannot Read From Extra Text File So Ignoring It\n", 1);
                                    bUseExtraFile = false;
                                }
                            }

                            if (bUseExtraFile) {
                                FILE *fp = fopen(ImgExtraText, "r");

                                if (fp != NULL) {
                                    bool bAddExtra = false;
                                    if (extraFileAge > 0) {
                                        struct stat buffer;
                                        if (stat(ImgExtraText, &buffer) == 0) {
                                            struct tm modifiedTime = *localtime(&buffer.st_mtime);

                                            time_t now = time(NULL);
                                            double ageInSeconds = difftime(now, mktime(&modifiedTime));
                                            sprintf(textBuffer, "Extra Text File (%s) Modified %f seconds ago ", ImgExtraText, ageInSeconds);
                                            displayDebugText(textBuffer, 1);
                                            if (ageInSeconds < extraFileAge) {
                                                displayDebugText("So Using It\n", 1);
                                                bAddExtra = true;
                                            } else {
                                                displayDebugText("So Ignoring\n", 1);
                                            }
                                        } else {
                                            displayDebugText("Stat Of Extra Text File Failed !\n", 1);
                                        }
                                    } else {
                                        displayDebugText("Extra Text File Age Disabled So Displaying Anyway\n", 1);
                                        bAddExtra = true;
                                    }
                                    if (bAddExtra) {
                                        char *line = NULL;
                                        size_t len = 0;
                                        while (getline(&line, &len, fp) != -1) {
                                            if (line[strlen(line)-1] == 10) {
                                                line[strlen(line)-1] = '\0';
                                            };
                                            cvText(pRgb, line, iTextX, iTextY + (iYOffset / currentBin), fontsize * SMALLFONTSIZE_MULTIPLIER, linewidth, linetype[linenumber], fontname[fontnumber], smallFontcolor, Image_type, outlinefont);
                                            iYOffset += iTextLineHeight;
                                        }
                                    }
                                    fclose(fp);
                                } else {
                                    displayDebugText("Failed To Open Extra Text File\n", 1);
                                }
                            }
                        } else {
                            displayDebugText("No Extra Text File Specified\n", 1);
                        }
                    }
#ifndef USE_NEW_EXPOSURE
                    sprintf(textBuffer, "  > Exposure value: %.0f µs (%.1f ms), actual ms=%.1lf\n", (float)actualExposureMicroseconds, (float)actualExposureMicroseconds/US_IN_MS, (double)actualExposureMicroseconds/US_IN_MS);
                    displayDebugText(textBuffer, 0);
#endif

#ifndef USE_HISTOGRAM
                    if (currentAutoExposure == 1)
                    {
                        // Retrieve the current Exposure for smooth transition to night time
                        // as long as auto-exposure is enabled during night time
                        currentExposure = actualExposureMicroseconds;
                    }
#endif

                    // Save the image
                    sprintf(textBuffer, "  > Saving image with time: %s\n", bufTime);   // ECC added "with time" portion
                    displayDebugText(textBuffer, 0);
                    if (!bSavingImg)
                    {
                        pthread_mutex_lock(&mtx_SaveImg);
                        pthread_cond_signal(&cond_SatrtSave);
                        pthread_mutex_unlock(&mtx_SaveImg);
                    }

                    if (asiNightAutoGain == 1 && dayOrNight == "NIGHT")
                    {
                        ASIGetControlValue(CamNum, ASI_GAIN, &actualGain, &bAuto);
                        sprintf(textBuffer, "Auto Gain value: %d\n", (int)actualGain);
                        displayDebugText(textBuffer, 0);
                        writeToLog(actualGain);
                    }

                    if (currentAutoExposure == 1)
                    {
#ifndef USE_HISTOGRAM
                        // printf("Auto Exposure value: %d ms\n", (int)round(actualExposureMicroseconds / US_IN_MS));  // ECC commented out - we output the exposure above
                        writeToLog(actualExposureMicroseconds);

                        if (dayOrNight == "DAY")
                        {
                            currentExposure = actualExposureMicroseconds;
                        }
#endif

                        // Delay applied before next exposure
                        if (dayOrNight == "NIGHT" && asiNightAutoExposure == 1 && actualExposureMicroseconds < (asiNightMaxExposure * US_IN_MS))
                        {
                            // If using auto-exposure and the actual exposure is less than the max,
                            // we still wait until we reach maxexposure, then wait for the delay period.
                            // This is important for a constant frame rate during timelapse generation.
                            // This doesn't apply during the day since we don't have a max time then.
                            int s = actualExposureMicroseconds - (asiNightMaxExposure * US_IN_MS); // to get to max
                            s += currentDelay * US_IN_MS;   // Add standard delay amount
                            sprintf(textBuffer,"  > Sleeping: %d ms\n", s / US_IN_MS);
                            displayDebugText(textBuffer, 0);
                            usleep(s);	// usleep() is in microseconds
                        }
                        else
                        {
                            sprintf(textBuffer,"  > Sleeping from auto exposure: %d ms (%.0f sec)\n", currentDelay, (float)currentDelay/US_IN_MS);
                            displayDebugText(textBuffer, 0);
                            usleep(currentDelay * US_IN_MS);
                        }
                    }
                    else
                    {
                        sprintf(textBuffer,"  > Sleeping from manual exposure: %d ms\n", currentDelay);    // ECC added
                        displayDebugText(textBuffer, 0);
                        usleep(currentDelay * US_IN_MS);
                    }
                    calculateDayOrNight(latitude, longitude, angle);

                } else {    // ECC added "else" portion.
                    sprintf(textBuffer,"  > ----- ASIGetVideoData() failed, retCode=%s, pRgb.step[0]=%d, pRgb.rows=%d, captureTimeout=%d\n", getRetCode(retCode), pRgb.step[0], pRgb.rows, captureTimeout);
                    displayDebugText(textBuffer, 0);

                    // Once the ASIGetVideoData() fails with a timeout, it seems to always fail,
                    // even with extremely large timeout values, so apparently, ASI_ERROR_TIMEOUT doesn't
                    // necessarily mean it's timing out.  I think it means the camera went away,
                    // so exit which will cause us to be restarted.
                    numErrors++; sleep(2);
                    if (numErrors >= maxErrors)
                    {
                        bMain = 0;  // get out of inner and outer loop
                        exitCode = 2;
                    }
                }
            }
            if (lastDayOrNight == "NIGHT")
            {
                endOfNight = true;
            }
#ifndef USE_NEW_EXPOSURE
            ASIStopVideoCapture(CamNum);
#endif
        }
    }
    ASICloseCamera(CamNum);

    if (bDisplay)
    {
        bDisplay = 0;
        pthread_join(thread_display, &retval);
    }

    if (bSaveRun)
    {
        bSaveRun = false;
        pthread_mutex_lock(&mtx_SaveImg);
        pthread_cond_signal(&cond_SatrtSave);
        pthread_mutex_unlock(&mtx_SaveImg);
        pthread_join(hthdSave, 0);
    }
    printf("main function over\n");
    return exitCode;
}
